@@ -32,38 +32,48 @@ static const char * const qgroup_cmd_group_usage[] = {
 	NULL
 };
 
-static int qgroup_assign(int assign, int argc, char **argv)
+static int _cmd_qgroup_assign(int assign, int argc, char **argv,
+		const char * const *usage_str)
 {
 	int ret = 0;
 	int fd;
-	int e;
 	int rescan = 0;
 	char *path;
 	struct btrfs_ioctl_qgroup_assign_args args;
 	DIR *dirstream = NULL;
 
-	while (1) {
-		enum { GETOPT_VAL_RESCAN = 256 };
-		static const struct option long_options[] = {
-			{ "rescan", no_argument, NULL, GETOPT_VAL_RESCAN },
-			{ NULL, 0, NULL, 0 }
-		};
-		int c = getopt_long(argc, argv, "", long_options, NULL);
+	if (assign) {
+		while (1) {
+			enum { GETOPT_VAL_RESCAN = 256, GETOPT_VAL_NO_RESCAN };
+			static const struct option long_options[] = {
+				{ "rescan", no_argument, NULL,
+					GETOPT_VAL_RESCAN },
+				{ "no-rescan", no_argument, NULL,
+					GETOPT_VAL_NO_RESCAN },
+				{ NULL, 0, NULL, 0 }
+			};
+			int c = getopt_long(argc, argv, "", long_options, NULL);
 
-		if (c < 0)
-			break;
-		switch (c) {
-		case GETOPT_VAL_RESCAN:
-			rescan = 1;
-			break;
-		default:
-			/* Usage printed by the caller */
-			return -1;
+			if (c < 0)
+				break;
+			switch (c) {
+			case GETOPT_VAL_RESCAN:
+				rescan = 1;
+				break;
+			case GETOPT_VAL_NO_RESCAN:
+				rescan = 0;
+				break;
+			default:
+				/* Usage printed by the caller */
+				return -1;
+			}
 		}
+	} else {
+		clean_args_no_options(argc, argv, usage_str);
 	}
 
 	if (check_argc_exact(argc - optind, 3))
-		return -1;
+		usage(usage_str);
 
 	memset(&args, 0, sizeof(args));
 	args.assign = assign;
@@ -76,7 +86,7 @@ static int qgroup_assign(int assign, int argc, char **argv)
 	 * FIXME src should accept subvol path
 	 */
 	if (btrfs_qgroup_level(args.src) >= btrfs_qgroup_level(args.dst)) {
-		fprintf(stderr, "ERROR: bad relation requested '%s'\n", path);
+		error("bad relation requested: %s", path);
 		return 1;
 	}
 	fd = btrfs_open_dir(path, &dirstream, 1);
@@ -84,10 +94,8 @@ static int qgroup_assign(int assign, int argc, char **argv)
 		return 1;
 
 	ret = ioctl(fd, BTRFS_IOC_QGROUP_ASSIGN, &args);
-	e = errno;
 	if (ret < 0) {
-		fprintf(stderr, "ERROR: unable to assign quota group: %s\n",
-			strerror(e));
+		error("unable to assign quota group: %s", strerror(errno));
 		close_file_or_dir(fd, dirstream);
 		return 1;
 	}
@@ -102,38 +110,38 @@ static int qgroup_assign(int assign, int argc, char **argv)
 	 */
 	if (ret > 0) {
 		if (rescan) {
-			struct btrfs_ioctl_quota_rescan_args args;
+			struct btrfs_ioctl_quota_rescan_args qargs;
 
 			printf("Quota data changed, rescan scheduled\n");
-			memset(&args, 0, sizeof(args));
-			ret = ioctl(fd, BTRFS_IOC_QUOTA_RESCAN, &args);
+			memset(&qargs, 0, sizeof(qargs));
+			ret = ioctl(fd, BTRFS_IOC_QUOTA_RESCAN, &qargs);
 			if (ret < 0)
-				fprintf(stderr,
-					"ERROR: quota rescan failed: %s\n",
+				error("quota rescan failed: %s",
 					strerror(errno));
 		} else {
-			printf("WARNING: quotas may be inconsistent, rescan needed\n");
+			warning("quotas may be inconsistent, rescan needed");
 		}
 	}
 	close_file_or_dir(fd, dirstream);
 	return ret;
 }
 
-static int qgroup_create(int create, int argc, char **argv)
+static int _cmd_qgroup_create(int create, int argc, char **argv)
 {
 	int ret = 0;
 	int fd;
 	int e;
-	char *path = argv[2];
+	char *path;
 	struct btrfs_ioctl_qgroup_create_args args;
 	DIR *dirstream = NULL;
 
-	if (check_argc_exact(argc, 3))
+	if (check_argc_exact(argc - optind, 2))
 		return -1;
 
 	memset(&args, 0, sizeof(args));
 	args.create = create;
-	args.qgroupid = parse_qgroupid(argv[1]);
+	args.qgroupid = parse_qgroupid(argv[optind]);
+	path = argv[optind + 1];
 
 	fd = btrfs_open_dir(path, &dirstream, 1);
 	if (fd < 0)
@@ -143,7 +151,7 @@ static int qgroup_create(int create, int argc, char **argv)
 	e = errno;
 	close_file_or_dir(fd, dirstream);
 	if (ret < 0) {
-		fprintf(stderr, "ERROR: unable to %s quota group: %s\n",
+		error("unable to %s quota group: %s",
 			create ? "create":"destroy", strerror(e));
 		return 1;
 	}
@@ -205,16 +213,13 @@ static const char * const cmd_qgroup_assign_usage[] = {
 	"Assign SRC as the child qgroup of DST",
 	"",
 	"--rescan       schedule qutoa rescan if needed",
-	"--no-rescan    ",
+	"--no-rescan    don't schedule quota rescan",
 	NULL
 };
 
 static int cmd_qgroup_assign(int argc, char **argv)
 {
-	int ret = qgroup_assign(1, argc, argv);
-	if (ret < 0)
-		usage(cmd_qgroup_assign_usage);
-	return ret;
+	return _cmd_qgroup_assign(1, argc, argv, cmd_qgroup_assign_usage);
 }
 
 static const char * const cmd_qgroup_remove_usage[] = {
@@ -225,10 +230,7 @@ static const char * const cmd_qgroup_remove_usage[] = {
 
 static int cmd_qgroup_remove(int argc, char **argv)
 {
-	int ret = qgroup_assign(0, argc, argv);
-	if (ret < 0)
-		usage(cmd_qgroup_remove_usage);
-	return ret;
+	return _cmd_qgroup_assign(0, argc, argv, cmd_qgroup_remove_usage);
 }
 
 static const char * const cmd_qgroup_create_usage[] = {
@@ -239,7 +241,12 @@ static const char * const cmd_qgroup_create_usage[] = {
 
 static int cmd_qgroup_create(int argc, char **argv)
 {
-	int ret = qgroup_create(1, argc, argv);
+	int ret;
+
+	clean_args_no_options(argc, argv, cmd_qgroup_create_usage);
+
+	ret = _cmd_qgroup_create(1, argc, argv);
+
 	if (ret < 0)
 		usage(cmd_qgroup_create_usage);
 	return ret;
@@ -253,7 +260,12 @@ static const char * const cmd_qgroup_destroy_usage[] = {
 
 static int cmd_qgroup_destroy(int argc, char **argv)
 {
-	int ret = qgroup_create(0, argc, argv);
+	int ret;
+
+	clean_args_no_options(argc, argv, cmd_qgroup_destroy_usage);
+
+	ret = _cmd_qgroup_create(0, argc, argv);
+
 	if (ret < 0)
 		usage(cmd_qgroup_destroy_usage);
 	return ret;
@@ -369,8 +381,7 @@ static int cmd_qgroup_show(int argc, char **argv)
 	e = errno;
 	close_file_or_dir(fd, dirstream);
 	if (ret < 0)
-		fprintf(stderr, "ERROR: can't list qgroups: %s\n",
-				strerror(e));
+		error("can't list qgroups: %s", strerror(e));
 
 	return !!ret;
 }
@@ -418,7 +429,7 @@ static int cmd_qgroup_limit(int argc, char **argv)
 		usage(cmd_qgroup_limit_usage);
 
 	if (!parse_limit(argv[optind], &size)) {
-		fprintf(stderr, "Invalid size argument given\n");
+		error("invalid size argument: %s", argv[optind]);
 		return 1;
 	}
 
@@ -439,12 +450,11 @@ static int cmd_qgroup_limit(int argc, char **argv)
 		path = argv[optind + 1];
 		ret = test_issubvolume(path);
 		if (ret < 0) {
-			fprintf(stderr, "ERROR: error accessing '%s'\n", path);
+			error("cannot access '%s': %s", path, strerror(-ret));
 			return 1;
 		}
 		if (!ret) {
-			fprintf(stderr, "ERROR: '%s' is not a subvolume\n",
-				path);
+			error("'%s' is not a subvolume", path);
 			return 1;
 		}
 		/*
@@ -465,8 +475,7 @@ static int cmd_qgroup_limit(int argc, char **argv)
 	e = errno;
 	close_file_or_dir(fd, dirstream);
 	if (ret < 0) {
-		fprintf(stderr, "ERROR: unable to limit requested quota group: "
-			"%s\n", strerror(e));
+		error("unable to limit requested quota group: %s", strerror(e));
 		return 1;
 	}
 	return 0;
