@@ -252,7 +252,6 @@ static int create_raid_groups(struct btrfs_trans_handle *trans,
 			      u64 metadata_profile, int mixed,
 			      struct mkfs_allocation *allocation)
 {
-	u64 num_devices = btrfs_super_num_devices(root->fs_info->super_copy);
 	int ret;
 
 	if (metadata_profile) {
@@ -271,7 +270,7 @@ static int create_raid_groups(struct btrfs_trans_handle *trans,
 		BUG_ON(ret);
 
 	}
-	if (!mixed && num_devices > 1 && data_profile) {
+	if (!mixed && data_profile) {
 		ret = create_one_raid_group(trans, root,
 					    BTRFS_BLOCK_GROUP_DATA |
 					    data_profile, allocation);
@@ -341,7 +340,7 @@ static void print_version(void)
 
 static u64 parse_profile(char *s)
 {
-	if (strcmp(s, "raid0") == 0) {
+	if (strcasecmp(s, "raid0") == 0) {
 		return BTRFS_BLOCK_GROUP_RAID0;
 	} else if (strcasecmp(s, "raid1") == 0) {
 		return BTRFS_BLOCK_GROUP_RAID1;
@@ -592,15 +591,14 @@ static int add_symbolic_link(struct btrfs_trans_handle *trans,
 			     u64 objectid, const char *path_name)
 {
 	int ret;
-	u64 sectorsize = root->sectorsize;
-	char *buf = malloc(sectorsize);
+	char buf[PATH_MAX];
 
-	ret = readlink(path_name, buf, sectorsize);
+	ret = readlink(path_name, buf, sizeof(buf));
 	if (ret <= 0) {
 		fprintf(stderr, "readlink failed for %s\n", path_name);
 		goto fail;
 	}
-	if (ret >= sectorsize) {
+	if (ret >= sizeof(buf)) {
 		fprintf(stderr, "symlink too long for %s\n", path_name);
 		ret = -1;
 		goto fail;
@@ -610,7 +608,6 @@ static int add_symbolic_link(struct btrfs_trans_handle *trans,
 	ret = btrfs_insert_inline_extent(trans, root, objectid, 0,
 					 buf, ret + 1);
 fail:
-	free(buf);
 	return ret;
 }
 
@@ -1032,16 +1029,15 @@ out:
  * This ignores symlinks with unreadable targets and subdirs that can't
  * be read.  It's a best-effort to give a rough estimate of the size of
  * a subdir.  It doesn't guarantee that prepopulating btrfs from this
- * tree won't still run out of space. 
- *
- * The rounding up to 4096 is questionable.  Previous code used du -B 4096.
+ * tree won't still run out of space.
  */
 static u64 global_total_size;
+static u64 fs_block_size;
 static int ftw_add_entry_size(const char *fpath, const struct stat *st,
 			      int type)
 {
 	if (type == FTW_F || type == FTW_D)
-		global_total_size += round_up(st->st_size, 4096);
+		global_total_size += round_up(st->st_size, fs_block_size);
 
 	return 0;
 }
@@ -1061,6 +1057,7 @@ static u64 size_sourcedir(char *dir_name, u64 sectorsize,
 			allocated_meta_size / default_chunk_size;
 
 	global_total_size = 0;
+	fs_block_size = sectorsize;
 	ret = ftw(dir_name, ftw_add_entry_size, 10);
 	dir_size = global_total_size;
 	if (ret < 0) {
@@ -1336,7 +1333,7 @@ out:
 	return ret;
 }
 
-int main(int ac, char **av)
+int main(int argc, char **argv)
 {
 	char *file;
 	struct btrfs_root *root;
@@ -1398,7 +1395,7 @@ int main(int ac, char **av)
 			{ NULL, 0, NULL, 0}
 		};
 
-		c = getopt_long(ac, av, "A:b:fl:n:s:m:d:L:O:r:U:VMKq",
+		c = getopt_long(argc, argv, "A:b:fl:n:s:m:d:L:O:r:U:VMKq",
 				long_options, NULL);
 		if (c < 0)
 			break;
@@ -1486,7 +1483,7 @@ int main(int ac, char **av)
 
 	sectorsize = max(sectorsize, (u32)sysconf(_SC_PAGESIZE));
 	saved_optind = optind;
-	dev_cnt = ac - optind;
+	dev_cnt = argc - optind;
 	if (dev_cnt == 0)
 		print_usage(1);
 
@@ -1510,16 +1507,16 @@ int main(int ac, char **av)
 	}
 
 	while (dev_cnt-- > 0) {
-		file = av[optind++];
+		file = argv[optind++];
 		if (is_block_device(file) == 1)
 			if (test_dev_for_mkfs(file, force_overwrite))
 				exit(1);
 	}
 
 	optind = saved_optind;
-	dev_cnt = ac - optind;
+	dev_cnt = argc - optind;
 
-	file = av[optind++];
+	file = argv[optind++];
 	ssd = is_ssd(file);
 
 	/*
@@ -1585,7 +1582,7 @@ int main(int ac, char **av)
 	for (i = saved_optind; i < saved_optind + dev_cnt; i++) {
 		char *path;
 
-		path = av[i];
+		path = argv[i];
 		ret = test_minimum_size(path, nodesize);
 		if (ret < 0) {
 			fprintf(stderr, "Failed to check size for '%s': %s\n",
@@ -1603,7 +1600,7 @@ int main(int ac, char **av)
 		}
 	}
 	ret = test_num_disk_vs_raid(metadata_profile, data_profile,
-			dev_cnt, mixed);
+			dev_cnt, mixed, ssd);
 	if (ret)
 		exit(1);
 
@@ -1731,7 +1728,7 @@ int main(int ac, char **av)
 		goto raid_groups;
 
 	while (dev_cnt-- > 0) {
-		file = av[optind++];
+		file = argv[optind++];
 
 		/*
 		 * open without O_EXCL so that the problem should not
