@@ -23,6 +23,7 @@
 #include "ctree.h"
 #include <dirent.h>
 #include <stdarg.h>
+#include "internal.h"
 
 #define BTRFS_MKFS_SYSTEM_GROUP_SIZE (4 * 1024 * 1024)
 #define BTRFS_MKFS_SMALL_VOLUME_SIZE (1024 * 1024 * 1024)
@@ -109,12 +110,16 @@ void btrfs_parse_features_to_string(char *buf, u64 flags);
 struct btrfs_mkfs_config {
 	char *label;
 	char *fs_uuid;
+	char *chunk_uuid;
 	u64 blocks[8];
 	u64 num_bytes;
 	u32 nodesize;
 	u32 sectorsize;
 	u32 stripesize;
 	u64 features;
+
+	/* Super bytenr after make_btrfs */
+	u64 super_bytenr;
 };
 
 int make_btrfs(int fd, struct btrfs_mkfs_config *cfg);
@@ -167,7 +172,7 @@ int test_dev_for_mkfs(char *file, int force_overwrite);
 int get_label_mounted(const char *mount_path, char *labelp);
 int get_label_unmounted(const char *dev, char *label);
 int test_num_disk_vs_raid(u64 metadata_profile, u64 data_profile,
-	u64 dev_cnt, int mixed);
+	u64 dev_cnt, int mixed, int ssd);
 int group_profile_max_safe_loss(u64 flags);
 int is_vol_small(char *file);
 int csum_tree_block(struct btrfs_root *root, struct extent_buffer *buf,
@@ -270,8 +275,10 @@ const char *get_argv0_buf(void);
 	"-t|--tbytes        show sizes in TiB, or TB with --si"
 
 unsigned int get_unit_mode_from_arg(int *argc, char *argv[], int df_mode);
+void clean_args_no_options(int argc, char *argv[], const char * const *usage);
 int string_is_numerical(const char *str);
 
+__attribute__ ((format (printf, 1, 2)))
 static inline void warning(const char *fmt, ...)
 {
 	va_list args;
@@ -283,6 +290,7 @@ static inline void warning(const char *fmt, ...)
 	fputc('\n', stderr);
 }
 
+__attribute__ ((format (printf, 1, 2)))
 static inline void error(const char *fmt, ...)
 {
 	va_list args;
@@ -294,6 +302,7 @@ static inline void error(const char *fmt, ...)
 	fputc('\n', stderr);
 }
 
+__attribute__ ((format (printf, 2, 3)))
 static inline int warning_on(int condition, const char *fmt, ...)
 {
 	va_list args;
@@ -310,6 +319,7 @@ static inline int warning_on(int condition, const char *fmt, ...)
 	return 1;
 }
 
+__attribute__ ((format (printf, 2, 3)))
 static inline int error_on(int condition, const char *fmt, ...)
 {
 	va_list args;
